// Generated by CoffeeScript 1.3.1
/*
Dialects are responsible for rendering an AST to a SQL string compatible with
a particular DBMS. They are rarely used directly, instead a query is usually bound
to an `engine <Engines>`_ that will delegate rendering to a private dialect
instance.
*/

var BaseDialect, MySQL, Postgres, prefixIfNotEmpty,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; };

prefixIfNotEmpty = function(prefix) {
  return function(node) {
    var children;
    children = this.renderNodeSet(node);
    if (children) {
      return prefix + children;
    } else {
      return '';
    }
  };
};

exports.BaseDialect = BaseDialect = (function() {

  BaseDialect.name = 'BaseDialect';

  function BaseDialect() {}

  /*
    A dialect that isn't specific to a particular DBMS, but used as a base for both
    MySQL and Postgres.
  */


  BaseDialect.prototype.render = function(node) {
    var name, type;
    type = node.__proto__;
    name = type.constructor.name;
    while (!this['render' + name]) {
      if ('Object' === name || !type) {
        throw new Error("Couldn't find a renderer for " + node.constructor.name);
      }
      type = type.__proto__;
      name = type.constructor.name;
    }
    return this['render' + name](node);
  };

  BaseDialect.prototype.renderString = function(s) {
    return s;
  };

  BaseDialect.prototype.renderNodeSet = function(set) {
    var _this = this;
    return set.nodes.map(function(n) {
      return _this.render(n);
    }).filter(function(n) {
      return n;
    }).join(set.glue);
  };

  BaseDialect.prototype.renderParenthesizedNodeSet = function(set) {
    return "(" + this.renderNodeSet(set) + ")";
  };

  BaseDialect.prototype.renderAlias = function(node) {
    return "" + (this.maybeParens(this.render(node.obj))) + " AS " + (this.render(node.alias));
  };

  BaseDialect.prototype.maybeParens = function(it) {
    if (/\s/.exec(it)) {
      return "(" + it + ")";
    } else {
      return it;
    }
  };

  BaseDialect.prototype.renderDistinct = function(set) {
    if (!set.enable) {
      return '';
    } else if (set.nodes.length) {
      return "DISTINCT(" + (this.renderNodeSet(set)) + ")";
    } else {
      return 'DISTINCT';
    }
  };

  BaseDialect.prototype.renderSelectProjectionSet = function(set) {
    if (!set.nodes.length) {
      return '*';
    } else {
      return this.renderNodeSet(set);
    }
  };

  BaseDialect.prototype.renderProjection = function(node) {
    return "" + (this.quote(node.source.ref())) + "." + (this.quote(node.field));
  };

  BaseDialect.prototype.quote = function(part) {
    if (/\./.exec(part)) {
      return "`" + part + "`";
    } else {
      return part;
    }
  };

  BaseDialect.prototype.renderSqlFunction = function(node) {
    return "" + (this.render(node.name)) + (this.render(node.arglist));
  };

  BaseDialect.prototype.renderValueNode = function(node) {
    return node.value;
  };

  BaseDialect.prototype.renderParameter = function(node) {
    return '?';
  };

  BaseDialect.prototype.renderSelect = prefixIfNotEmpty('SELECT ');

  BaseDialect.prototype.renderUpdate = prefixIfNotEmpty('UPDATE ');

  BaseDialect.prototype.renderInsert = prefixIfNotEmpty('INSERT INTO ');

  BaseDialect.prototype.renderInsertData = prefixIfNotEmpty('VALUES ');

  BaseDialect.prototype.renderDelete = prefixIfNotEmpty('DELETE ');

  BaseDialect.prototype.renderUpdateSet = prefixIfNotEmpty('SET ');

  BaseDialect.prototype.renderRelationSet = prefixIfNotEmpty('FROM ');

  BaseDialect.prototype.renderWhere = prefixIfNotEmpty('WHERE ');

  BaseDialect.prototype.renderGroupBy = prefixIfNotEmpty('GROUP BY ');

  BaseDialect.prototype.renderOrderBySet = prefixIfNotEmpty('ORDER BY ');

  BaseDialect.prototype.renderReturning = prefixIfNotEmpty('RETURNING ');

  BaseDialect.prototype.renderLimit = function(node) {
    if (node.value) {
      return "LIMIT " + node.value;
    } else {
      return "";
    }
  };

  BaseDialect.prototype.renderOffset = function(node) {
    if (node.value) {
      return "OFFSET " + node.value;
    } else {
      return "";
    }
  };

  BaseDialect.prototype.renderBinary = function(node) {
    return this.render(node.left) + ' ' + this.renderOp(node.op) + ' ' + this.render(node.right);
  };

  BaseDialect.prototype.renderOp = function(op) {
    switch (op.toLowerCase()) {
      case 'ne':
      case '!=':
      case '<>':
        return '!=';
      case 'eq':
      case '=':
        return '=';
      case 'lt':
      case '<':
        return '<';
      case 'gt':
      case '>':
        return '>';
      case 'lte':
      case '<=':
        return '<=';
      case 'gte':
      case '>=':
        return '>=';
      case 'like':
        return 'LIKE';
      case 'ilike':
        return 'ILIKE';
      case 'in':
        return 'IN';
      case 'is':
        return 'IS';
      default:
        throw new Error("Unsupported comparison operator: " + op);
    }
  };

  return BaseDialect;

})();

exports.Postgres = Postgres = (function(_super) {
  var Delete, Insert, Select, Update, _ref;

  __extends(Postgres, _super);

  /* Specialization of BaseDialect for Postgres
  */


  Postgres.name = 'Postgres';

  function Postgres() {
    return Postgres.__super__.constructor.apply(this, arguments);
  }

  _ref = require('./nodes'), Select = _ref.Select, Update = _ref.Update, Delete = _ref.Delete, Insert = _ref.Insert;

  Postgres.prototype.render = function(node) {
    var _ref1;
    if ((_ref1 = node.constructor) === Select || _ref1 === Update || _ref1 === Delete || _ref1 === Insert) {
      this.paramCount = 1;
    }
    return Postgres.__super__.render.call(this, node);
  };

  Postgres.prototype.renderParameter = function(node) {
    return "$" + (this.paramCount++);
  };

  Postgres.prototype.renderOp = function(op) {
    switch (op.toLowerCase()) {
      case 'hasKey':
        return '?';
      case '->':
        return '->';
      default:
        return Postgres.__super__.renderOp.call(this, op);
    }
  };

  return Postgres;

})(BaseDialect);

exports.MySQL = MySQL = (function(_super) {

  __extends(MySQL, _super);

  MySQL.name = 'MySQL';

  function MySQL() {
    return MySQL.__super__.constructor.apply(this, arguments);
  }

  /*
    Specialization of BaseDialect for MySQL
  */


  MySQL.prototype.renderParameter = function(node) {
    return '?';
  };

  return MySQL;

})(BaseDialect);
