// Generated by CoffeeScript 1.3.1
var createTxProxy, dialects, fakeEngine, mysql, postgres, queries, transaction;

queries = require('./queries');

dialects = require('./dialects');

postgres = function(config) {
  /*
    Create a new Postgres engine using a parameter compatible with
    `https://github.com/brianc/node-postgres/wiki/pg#method-connect <require('pg').connect(config)>`_.
  
    This will ``require('pg')`` and attempt to use the 'native' interface if it's
    available.
  */

  var dialect, engine, pg;
  pg = require('pg');
  if (pg["native"]) {
    pg = pg["native"];
  }
  dialect = new dialects.Postgres;
  engine = {
    name: "postgres",
    params: config,
    render: dialect.render.bind(dialect),
    connect: pg.connect.bind(pg, config),
    destroy: pg.end
  };
  engine.transaction = transaction.bind(engine);
  queries.mixinFactoryMethods(engine);
  return engine;
};

mysql = function(opts) {
  /*
    Create a new MySQL engine using an object that is compatible with
    ``require('mysql').create{Connection,Client}(opts)``.
  
    Additionally, you can specify extra options for ``generic-pool`` by
    including them as an object in ``opts.pool``. The ``create`` and
    ``destroy`` pool functions will be created for you.
    
    This will ``require('mysql-compat')`` which you **must** install separately
    for this function to work.
  */

  var dialect, engine;
  mysql = require('mysql-compat');
  dialect = new dialects.MySQL;
  engine = {
    name: "mysql",
    params: opts,
    render: dialect.render.bind(dialect),
    connect: mysql.connect.bind(null, opts),
    destroy: mysql.end
  };
  engine.transaction = transaction.bind(engine);
  queries.mixinFactoryMethods(engine);
  return engine;
};

fakeEngine = function() {
  /*
    Create a no-op engine that simply returns the compiled SQL and parameter
    array to the result callback. This will be the default until you over-ride
    with ``gesundheit.defaultEngine = myAppEngine``.
  */

  var bd, engine, passthroughClient;
  bd = new dialects.BaseDialect;
  engine = {
    params: {},
    render: bd.render.bind(bd),
    connect: function(cb) {
      return cb(null, passthroughClient);
    }
  };
  passthroughClient = {
    engine: engine,
    query: function(sql, params, cb) {
      if (!cb) {
        return;
      }
      return process.nextTick(cb.bind(null, null, [sql, params]));
    }
  };
  return engine;
};

transaction = function(cb) {
  var _this = this;
  return this.connect(function(err, conn) {
    var tx;
    if (err) {
      return cb(err);
    }
    tx = createTxProxy.call(_this, conn);
    return tx.begin(function(err) {
      return cb(err, tx);
    });
  });
};

createTxProxy = function(conn) {
  var proxy;
  proxy = {
    connection: conn,
    begin: function(cb) {
      conn.pauseDrain();
      return conn.query('BEGIN', function(err) {
        return cb(err);
      });
    },
    commit: function(cb) {
      return conn.query('COMMIT', function(err) {
        conn.resumeDrain();
        return cb(err);
      });
    },
    rollback: function(cb) {
      return conn.query('ROLLBACK', function(err) {
        conn.resumeDrain();
        return cb(err);
      });
    },
    query: conn.query.bind(conn),
    connect: function(cb) {
      return cb(null, conn);
    },
    render: this.render.bind(this)
  };
  queries.mixinFactoryMethods(proxy);
  return proxy;
};

module.exports = {
  mysql: mysql,
  postgres: postgres,
  fakeEngine: fakeEngine
};
