// Generated by CoffeeScript 1.3.1
/*
These are the classes that represent nodes in the AST for a SQL statement.
Application code should very rarely have to deal with these classes directly;
Instead, the APIs exposed by the various query manager classes are intended to
cover the majority of use-cases. However, in the spirit of "making hard things
possible", the various AST nodes can be constructed and assembled manually if you
so desire.
*/

var Alias, And, Binary, CONST_NODES, ComparableMixin, Delete, Distinct, FixedNamedNodeSet, FixedNodeSet, GroupBy, Insert, InsertData, JOIN_TYPES, Join, JoinType, Limit, Node, NodeSet, Offset, Or, OrderBy, OrderBySet, Parameter, ParenthesizedNodeSet, Projection, ProjectionSet, Relation, RelationSet, Returning, Select, SelectProjectionSet, SqlFunction, Tuple, Update, UpdateSet, ValueNode, Where, binaryOp, copy, getAlias, k, name, sqlFunction, text, toParam, toRelation, v, _i, _j, _len, _len1, _ref, _ref1, _ref2,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; },
  __slice = [].slice;

Node = (function() {
  /* (Empty) base Node class
  */

  Node.name = 'Node';

  function Node() {}

  return Node;

})();

ValueNode = (function(_super) {

  __extends(ValueNode, _super);

  ValueNode.name = 'ValueNode';

  /* A ValueNode is a literal string that should be printed unescaped.
  */


  function ValueNode(value) {
    this.value = value;
  }

  ValueNode.prototype.copy = function() {
    return new this.constructor(this.value);
  };

  return ValueNode;

})(Node);

CONST_NODES = {};

_ref = ['DEFAULT', 'NULL', 'IS_NULL', 'IS_NOT_NULL'];
for (_i = 0, _len = _ref.length; _i < _len; _i++) {
  name = _ref[_i];
  CONST_NODES[name] = new ValueNode(name.replace('_', ' '));
}

JoinType = (function(_super) {

  __extends(JoinType, _super);

  JoinType.name = 'JoinType';

  function JoinType() {
    return JoinType.__super__.constructor.apply(this, arguments);
  }

  return JoinType;

})(ValueNode);

JOIN_TYPES = {};

_ref1 = ['LEFT', 'RIGHT', 'INNER', 'LEFT_OUTER', 'RIGHT_OUTER', 'FULL_OUTER', 'NATURAL', 'CROSS'];
for (_j = 0, _len1 = _ref1.length; _j < _len1; _j++) {
  name = _ref1[_j];
  JOIN_TYPES[name] = new ValueNode(name.replace('_', ' '));
}

NodeSet = (function(_super) {

  __extends(NodeSet, _super);

  NodeSet.name = 'NodeSet';

  /* A set of nodes joined together by ``@glue``
  */


  function NodeSet(nodes, glue) {
    this.nodes = nodes != null ? nodes : [];
    if (glue == null) {
      glue = ' ';
    }
    /*
        :param @nodes: A list of child nodes.
        :param glue: A string that will be used to join the nodes when rendering
    */

    this.glue || (this.glue = glue);
  }

  NodeSet.prototype.copy = function() {
    /*
        Make a deep copy of this node and it's children
    */

    var c;
    c = new this.constructor;
    c.nodes = copy(this.nodes);
    return c;
  };

  NodeSet.prototype.addNode = function(node) {
    /* Add a new Node to the end of this set
    */
    return this.nodes.push(node);
  };

  NodeSet.prototype.params = function() {
    /*
        Recurse over nested NodeSet instances, collecting parameter values.
    */

    var node, params, _k, _len2, _ref2;
    params = [];
    _ref2 = this.nodes;
    for (_k = 0, _len2 = _ref2.length; _k < _len2; _k++) {
      node = _ref2[_k];
      if (node.params != null) {
        params = params.concat(node.params());
      } else if (node.constructor === Parameter) {
        params.push(node.value);
      }
    }
    return params;
  };

  return NodeSet;

})(Node);

FixedNodeSet = (function(_super) {

  __extends(FixedNodeSet, _super);

  FixedNodeSet.name = 'FixedNodeSet';

  /* A NodeSet that disables the ``addNode`` method after construction.
  */


  function FixedNodeSet(nodes, glue) {
    FixedNodeSet.__super__.constructor.call(this, nodes, glue);
    this.addNode = null;
  }

  return FixedNodeSet;

})(NodeSet);

FixedNamedNodeSet = (function(_super) {

  __extends(FixedNamedNodeSet, _super);

  FixedNamedNodeSet.name = 'FixedNamedNodeSet';

  /*
    A FixedNodeSet that instantiates a set of nodes defined by the class member
    ``@structure`` when it it instantiated.
  
    See :class:`nodes::Select` for example.
  */


  function FixedNamedNodeSet() {
    var k, nodes, type;
    nodes = (function() {
      var _k, _len2, _ref2, _ref3, _results;
      _ref2 = this.constructor.structure;
      _results = [];
      for (_k = 0, _len2 = _ref2.length; _k < _len2; _k++) {
        _ref3 = _ref2[_k], k = _ref3[0], type = _ref3[1];
        _results.push(this[k] = new type);
      }
      return _results;
    }).call(this);
    FixedNamedNodeSet.__super__.constructor.call(this, nodes);
  }

  FixedNamedNodeSet.prototype.copy = function() {
    var c, i, k, _, _ref2, _ref3;
    c = FixedNamedNodeSet.__super__.copy.call(this);
    _ref2 = this.constructor.structure;
    for (i in _ref2) {
      _ref3 = _ref2[i], k = _ref3[0], _ = _ref3[1];
      c[k] = c.nodes[i];
    }
    return c;
  };

  return FixedNamedNodeSet;

})(FixedNodeSet);

SqlFunction = SqlFunction = (function(_super) {

  __extends(SqlFunction, _super);

  SqlFunction.name = 'SqlFunction';

  /* Includes :class:`nodes::ComparableMixin`
  */


  function SqlFunction(name, arglist) {
    this.name = name;
    this.arglist = arglist;
  }

  SqlFunction.prototype.copy = function() {
    return new this.constructor(this.name, copy(this.arglist));
  };

  return SqlFunction;

})(Node);

Alias = (function(_super) {

  __extends(Alias, _super);

  Alias.name = 'Alias';

  /*
    Example::
  
       table = new Relation('my_table_with_a_long_name')
       alias = new Alias(table, 'mtwaln')
  */


  function Alias(obj, alias) {
    this.obj = obj;
    this.alias = alias;
  }

  Alias.prototype.copy = function() {
    return new this.constructor(copy(this.obj), this.alias);
  };

  Alias.prototype.project = function(name) {
    return new Projection(this, name);
  };

  Alias.prototype.field = function() {
    return this.project(arguments);
  };

  Alias.prototype.ref = function() {
    return this.alias;
  };

  return Alias;

})(Node);

exports.Parameter = Parameter = (function(_super) {

  __extends(Parameter, _super);

  /*
    Like a ValueNode, but will render as a bound parameter place-holder
    (e.g. '?' for MySQL) and it's value will be collected by
    :meth:`nodes::NodeSet.params`
  */


  Parameter.name = 'Parameter';

  function Parameter() {
    return Parameter.__super__.constructor.apply(this, arguments);
  }

  return Parameter;

})(ValueNode);

exports.Relation = Relation = (function(_super) {

  __extends(Relation, _super);

  Relation.name = 'Relation';

  function Relation() {
    return Relation.__super__.constructor.apply(this, arguments);
  }

  /*
    A relation node represents a table name in a statement.
  */


  Relation.prototype.ref = function() {
    /*
        Return the table name. This is a common interface with :class:`nodes:Alias`.
    */
    return this.value;
  };

  Relation.prototype.project = function(field) {
    /* Return a new :class:`nodes::Projection` of `field` from this table.
    */
    return new Projection(this, field);
  };

  Relation.prototype.field = function(field) {
    /* An alias for :meth:`nodes::Relation.project`.
    */
    return this.project(field);
  };

  Relation.prototype.copy = function() {
    return new this.constructor(this.value);
  };

  return Relation;

})(ValueNode);

exports.Limit = Limit = (function(_super) {

  __extends(Limit, _super);

  Limit.name = 'Limit';

  function Limit() {
    return Limit.__super__.constructor.apply(this, arguments);
  }

  return Limit;

})(ValueNode);

exports.Offset = Offset = (function(_super) {

  __extends(Offset, _super);

  Offset.name = 'Offset';

  function Offset() {
    return Offset.__super__.constructor.apply(this, arguments);
  }

  return Offset;

})(ValueNode);

Binary = (function(_super) {

  __extends(Binary, _super);

  Binary.name = 'Binary';

  function Binary(left, op, right) {
    this.left = left;
    this.op = op;
    this.right = right;
    Binary.__super__.constructor.call(this, [this.left, this.op, this.right], ' ');
  }

  Binary.prototype.copy = function() {
    return new this.constructor(copy(this.left), this.op, copy(this.right));
  };

  return Binary;

})(FixedNodeSet);

ParenthesizedNodeSet = (function(_super) {

  __extends(ParenthesizedNodeSet, _super);

  /* A NodeSet wrapped in parenthesis.
  */


  ParenthesizedNodeSet.name = 'ParenthesizedNodeSet';

  function ParenthesizedNodeSet() {
    return ParenthesizedNodeSet.__super__.constructor.apply(this, arguments);
  }

  return ParenthesizedNodeSet;

})(NodeSet);

Tuple = (function(_super) {

  __extends(Tuple, _super);

  Tuple.name = 'Tuple';

  function Tuple() {
    return Tuple.__super__.constructor.apply(this, arguments);
  }

  /* A tuple node. e.g. (col1, col2)
  */


  Tuple.prototype.glue = ', ';

  return Tuple;

})(ParenthesizedNodeSet);

ProjectionSet = (function(_super) {

  __extends(ProjectionSet, _super);

  ProjectionSet.name = 'ProjectionSet';

  function ProjectionSet() {
    return ProjectionSet.__super__.constructor.apply(this, arguments);
  }

  /* The list of projected columns in a query
  */


  ProjectionSet.prototype.glue = ', ';

  return ProjectionSet;

})(NodeSet);

Returning = (function(_super) {

  __extends(Returning, _super);

  Returning.name = 'Returning';

  function Returning() {
    return Returning.__super__.constructor.apply(this, arguments);
  }

  return Returning;

})(ProjectionSet);

Distinct = (function(_super) {

  __extends(Distinct, _super);

  Distinct.name = 'Distinct';

  function Distinct(enable) {
    this.enable = enable != null ? enable : false;
    Distinct.__super__.constructor.apply(this, arguments);
  }

  return Distinct;

})(ProjectionSet);

SelectProjectionSet = (function(_super) {

  __extends(SelectProjectionSet, _super);

  SelectProjectionSet.name = 'SelectProjectionSet';

  function SelectProjectionSet() {
    return SelectProjectionSet.__super__.constructor.apply(this, arguments);
  }

  SelectProjectionSet.prototype.prune = function(predicate) {
    /*
        Recurse over child nodes, removing all Projection nodes that match the
        predicate.
    */

    var node, orig, _results;
    orig = this.nodes;
    this.nodes = [];
    _results = [];
    for (node in orig) {
      if (node instanceof Projection) {
        if (!predicate(node)) {
          _results.push(this.nodes.push(node));
        } else {
          _results.push(void 0);
        }
      } else if (node instanceof Alias) {
        if (!pred(node.obj)) {
          _results.push(this.nodes.push(node));
        } else {
          _results.push(void 0);
        }
      } else {
        _results.push(void 0);
      }
    }
    return _results;
  };

  return SelectProjectionSet;

})(ProjectionSet);

Projection = (function(_super) {

  __extends(Projection, _super);

  Projection.name = 'Projection';

  /*
    Includes :class:`nodes::ComparableMixin`
  */


  function Projection(source, field) {
    this.source = source;
    this.field = field;
    Projection.__super__.constructor.call(this, [this.source, this.field], '.');
  }

  Projection.prototype.copy = function() {
    return new this.constructor(copy(this.source), this.field);
  };

  return Projection;

})(FixedNodeSet);

exports.RelationSet = RelationSet = (function(_super) {

  __extends(RelationSet, _super);

  RelationSet.name = 'RelationSet';

  function RelationSet() {
    return RelationSet.__super__.constructor.apply(this, arguments);
  }

  /*
    Manages a set of relation and exposes methods to find them by alias.
  */


  RelationSet.prototype.start = function(first) {
    this.first = first;
    this.byName = {};
    this.nodes.unshift(this.first);
    this.byName[this.first.ref()] = this.active = this.first;
    return delete this.start;
  };

  RelationSet.prototype.registerName = function(node) {
    return this.byName[node.ref()] = node;
  };

  RelationSet.prototype.copy = function() {
    var c, first, node, rel, _k, _len2, _ref2;
    c = RelationSet.__super__.copy.call(this);
    if (first = c.nodes.shift()) {
      c.start(first);
    }
    _ref2 = c.nodes;
    for (_k = 0, _len2 = _ref2.length; _k < _len2; _k++) {
      node = _ref2[_k];
      rel = node.relation || node;
      c.registerName(rel);
    }
    if (first) {
      c.active = c.get(this.active);
    }
    return c;
  };

  RelationSet.prototype.get = function(name, strict) {
    var rel;
    if (strict == null) {
      strict = true;
    }
    if ('string' !== typeof name) {
      name = name.ref();
    }
    rel = this.byName[name];
    if (!rel && strict) {
      throw new Error("No such relation " + name + " in " + (Object.keys(this.byName)));
    }
    return rel;
  };

  RelationSet.prototype["switch"] = function(name) {
    return this.active = this.get(name);
  };

  return RelationSet;

})(NodeSet);

exports.Join = Join = (function(_super) {

  __extends(Join, _super);

  Join.name = 'Join';

  function Join(type, relation, on) {
    var nodes;
    this.type = type;
    this.relation = relation;
    this.on = on;
    nodes = [this.type, 'JOIN', this.relation];
    if (this.on) {
      nodes.push('ON', this.on);
    }
    Join.__super__.constructor.call(this, nodes);
  }

  Join.prototype.copy = function() {
    return new this.constructor(this.type, copy(this.relation), copy(this.on));
  };

  return Join;

})(FixedNodeSet);

Where = (function(_super) {

  __extends(Where, _super);

  Where.name = 'Where';

  function Where() {
    return Where.__super__.constructor.apply(this, arguments);
  }

  Where.prototype.glue = ' AND ';

  return Where;

})(NodeSet);

Or = (function(_super) {

  __extends(Or, _super);

  Or.name = 'Or';

  function Or() {
    return Or.__super__.constructor.apply(this, arguments);
  }

  Or.prototype.glue = ' OR ';

  return Or;

})(ParenthesizedNodeSet);

exports.And = And = (function(_super) {

  __extends(And, _super);

  And.name = 'And';

  function And() {
    return And.__super__.constructor.apply(this, arguments);
  }

  And.prototype.glue = ' AND ';

  return And;

})(ParenthesizedNodeSet);

exports.GroupBy = GroupBy = (function(_super) {

  __extends(GroupBy, _super);

  GroupBy.name = 'GroupBy';

  function GroupBy(gs) {
    GroupBy.__super__.constructor.call(this, gs, ', ');
  }

  return GroupBy;

})(NodeSet);

exports.OrderBySet = OrderBySet = (function(_super) {

  __extends(OrderBySet, _super);

  OrderBySet.name = 'OrderBySet';

  function OrderBySet(os) {
    OrderBySet.__super__.constructor.call(this, os, ', ');
  }

  return OrderBySet;

})(NodeSet);

exports.OrderBy = OrderBy = (function(_super) {

  __extends(OrderBy, _super);

  OrderBy.name = 'OrderBy';

  function OrderBy(projection, direction) {
    OrderBy.__super__.constructor.call(this, [projection, direction]);
  }

  return OrderBy;

})(FixedNodeSet);

exports.UpdateSet = UpdateSet = (function(_super) {

  __extends(UpdateSet, _super);

  UpdateSet.name = 'UpdateSet';

  function UpdateSet(nodes) {
    UpdateSet.__super__.constructor.call(this, nodes, ', ');
  }

  return UpdateSet;

})(NodeSet);

exports.Select = Select = (function(_super) {

  __extends(Select, _super);

  Select.name = 'Select';

  Select.structure = [['distinct', Distinct], ['projections', SelectProjectionSet], ['relations', RelationSet], ['where', Where], ['groupBy', GroupBy], ['orderBy', OrderBySet], ['limit', Limit], ['offset', Offset]];

  function Select(rel) {
    Select.__super__.constructor.call(this);
    if (rel) {
      this.relations.start(rel);
    }
  }

  return Select;

})(FixedNamedNodeSet);

Update = (function(_super) {

  __extends(Update, _super);

  Update.name = 'Update';

  Update.structure = [['relation', Relation], ['updates', UpdateSet], ['orderBy', OrderBySet], ['limit', Limit], ['fromList', RelationSet], ['where', Where], ['returning', Returning]];

  function Update(rel) {
    Update.__super__.constructor.call(this);
    if (rel) {
      this.relation = this.nodes[0] = rel;
    }
  }

  return Update;

})(FixedNamedNodeSet);

InsertData = (function(_super) {

  __extends(InsertData, _super);

  InsertData.name = 'InsertData';

  function InsertData() {
    return InsertData.__super__.constructor.apply(this, arguments);
  }

  InsertData.prototype.glue = ', ';

  return InsertData;

})(NodeSet);

Insert = (function(_super) {

  __extends(Insert, _super);

  Insert.name = 'Insert';

  Insert.structure = [['relation', Relation], ['columns', Tuple], ['source', InsertData], ['returning', Returning]];

  function Insert(rel) {
    Insert.__super__.constructor.call(this);
    if (rel) {
      this.relation = this.nodes[0] = rel;
    }
  }

  Insert.prototype.addRow = function(row) {
    if (this.source.constructor === Select) {
      throw new Error("Cannot add rows when inserting from a SELECT");
    }
    if (row.constructor === Array) {
      return this.addRowArray(row);
    } else {
      return this.addRowObject(row);
    }
  };

  Insert.prototype.addRowArray = function(row) {
    var count, fields, n, params, v;
    if (!(count = this.columns.nodes.length)) {
      throw new Error("Must set column list before inserting arrays");
    }
    if (row.length !== count) {
      fields = (function() {
        var _k, _len2, _ref2, _results;
        _ref2 = this.columns.nodes;
        _results = [];
        for (_k = 0, _len2 = _ref2.length; _k < _len2; _k++) {
          n = _ref2[_k];
          _results.push(n);
        }
        return _results;
      }).call(this);
      throw new Error("Wrong number of values in array, expected " + fields);
    }
    params = (function() {
      var _k, _len2, _results;
      _results = [];
      for (_k = 0, _len2 = row.length; _k < _len2; _k++) {
        v = row[_k];
        if (v instanceof Node) {
          _results.push(v);
        } else {
          _results.push(new Parameter(v));
        }
      }
      return _results;
    })();
    return this.source.addNode(new Tuple(params));
  };

  Insert.prototype.addRowObject = function(row) {
    /*
        Add a row from an object. This will set the column list of the query if it
        isn't set yet. If it _is_ set, then only keys matching the existing column
        list will be inserted.
    */

    var array, f;
    array = (function() {
      var _k, _len2, _ref2, _results;
      _ref2 = this.columns.nodes;
      _results = [];
      for (_k = 0, _len2 = _ref2.length; _k < _len2; _k++) {
        f = _ref2[_k];
        if ((row[f] != null) || row[f] === null) {
          _results.push(row[f]);
        } else {
          _results.push(CONST_NODES.DEFAULT);
        }
      }
      return _results;
    }).call(this);
    return this.addRowArray(array);
  };

  return Insert;

})(FixedNamedNodeSet);

Delete = (function(_super) {

  __extends(Delete, _super);

  Delete.name = 'Delete';

  Delete.structure = [['relations', RelationSet], ['where', Where], ['orderBy', OrderBySet], ['limit', Limit]];

  function Delete(rel) {
    Delete.__super__.constructor.call(this);
    if (rel) {
      this.relations.start(rel);
    }
  }

  return Delete;

})(FixedNamedNodeSet);

ComparableMixin = (function() {

  ComparableMixin.name = 'ComparableMixin';

  function ComparableMixin() {}

  /*
    A mixin that adds comparison methods to a class. Each of these comparison
    methods will yield a new AST node comparing the invocant to the argument.
  */


  ComparableMixin.prototype.eq = function(other) {
    /* ``this = other``
    */
    return new Binary(this, '=', toParam(other));
  };

  ComparableMixin.prototype.ne = function(other) {
    /* ``this <> other``
    */
    return new Binary(this, '<>', toParam(other));
  };

  ComparableMixin.prototype.gt = function(other) {
    /* ``this > other``
    */
    return new Binary(this, '>', toParam(other));
  };

  ComparableMixin.prototype.lt = function(other) {
    /* ``this < other``
    */
    return new Binary(this, '<', toParam(other));
  };

  ComparableMixin.prototype.lte = function(other) {
    /* ``this <= other``
    */
    return new Binary(this, '<=', toParam(other));
  };

  ComparableMixin.prototype.gte = function(other) {
    /* ``this >= other``
    */
    return new Binary(this, '>=', toParam(other));
  };

  ComparableMixin.prototype.compare = function(op, other) {
    /* ``this op other``
    */
    return new Binary(this, op, toParam(other));
  };

  return ComparableMixin;

})();

_ref2 = ComparableMixin.prototype;
for (k in _ref2) {
  v = _ref2[k];
  Projection.prototype[k] = v;
  SqlFunction.prototype[k] = v;
}

toParam = function(it) {
  /*
    Return a Node that can be used as a parameter.
  
      * SelectQuery instances will be treated as un-named sub queries,
      * Node instances will be returned unchanged.
      * Arrays will be turned into a :class:`nodes::Tuple` instance.
    
    All other types will be wrapped in a :class:`nodes::Parameter` instance.
  */

  var SelectQuery;
  SelectQuery = require('./queries/select');
  if ((it != null ? it.constructor : void 0) === SelectQuery) {
    return new Tuple([it.q]);
  } else if (it instanceof Node) {
    return it;
  } else if (Array.isArray(it)) {
    return new Tuple(it.map(toParam));
  } else {
    return new Parameter(it);
  }
};

toRelation = toRelation = function(it) {
  /*
    Transform ``it`` into a :class:`nodes::Relation` instance.
  
    This accepts `strings, `Relation`` and ``Alias`` instances, and objects with
    a single key-value pair, which will be turned into an ``Alias`` instance.
    
    Examples::
    
       toRelation('some_table') # new Relation('some_table')
       toRelation(st: 'some_table') # new Alias(Relation('some_table'), 'st')
       toRelation(new Relation('some_table')) # returns same instance
       toRelation(new Alias(new Relation('some_table'), 'al') # returns same instance
   
    **Throws Errors** if the input is not valid.
  */

  var alias;
  switch (it.constructor) {
    case Relation:
    case Alias:
      return it;
    case String:
      return new Relation(it);
    case Object:
      if (alias = getAlias(it)) {
        return new Alias(new Relation(it[alias]), alias);
      } else {
        throw new Error("Can't make relation out of " + it);
      }
      break;
    default:
      throw new Error("Can't make relation out of " + it);
  }
};

sqlFunction = function(name, args) {
  /*
    Create a new SQL function call node. For example::
  
        count = sqlFunction('count', new ValueNode('*'))
  */
  return new SqlFunction(name, new Tuple(args));
};

getAlias = function(o) {
  /*
    Check if ``o`` is an object literal representing an alias, and return the
    alias name if it is.
  */

  var keys;
  if ('object' === typeof o) {
    keys = Object.keys(o);
    if (keys.length === 1) {
      return keys[0];
    }
  }
  return null;
};

text = function() {
  var node, params, rawSQL;
  rawSQL = arguments[0], params = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
  /*
    Construct a node with a raw SQL string and (optionally) parameters.
  
    Parameter arguments are assumed to be values for placeholders in the raw
    string. Be careful: the number and types of these parameters will **not**
    be checked, so it is very easy to create invalid statements with this.
  */

  node = new ValueNode(rawSQL);
  node.params = function() {
    return params;
  };
  return node;
};

binaryOp = function(left, op, right) {
  /*
    Create a new :class:`nodes::Binary` node::
  
      binaryOp('hstore_column', '->', toParam(y))
      # hstore_column -> ?
  
    .. seealso::
  
      :meth:`queries/select::SelectQuery.project`
         Returns :class:`nodes::Projection` objects that have comparison methods
         from :class:`nodes::ComparableMixin`.
  */
  return new Binary(left, op, right);
};

module.exports = {
  CONST_NODES: CONST_NODES,
  JOIN_TYPES: JOIN_TYPES,
  binaryOp: binaryOp,
  getAlias: getAlias,
  sqlFunction: sqlFunction,
  text: text,
  toRelation: toRelation,
  toParam: toParam,
  Alias: Alias,
  And: And,
  Or: Or,
  Join: Join,
  OrderBy: OrderBy,
  Tuple: Tuple,
  FixedNamedNodeSet: FixedNamedNodeSet,
  Select: Select,
  Update: Update,
  Insert: Insert,
  Delete: Delete
};

copy = function(it) {
  /*
    Return a deep copy of ``it``.
  */

  var c, k, v, _results;
  if (!it) {
    return it;
  }
  switch (it.constructor) {
    case String:
    case Number:
    case Boolean:
      return it;
    case Array:
      return it.map(copy);
    case Object:
      c = {};
      _results = [];
      for (k in it) {
        v = it[k];
        _results.push(c[k] = copy(v));
      }
      return _results;
      break;
    default:
      if (it.copy != null) {
        return it.copy();
      } else {
        throw new Error("Don't know how to copy " + it);
      }
  }
};
