// Generated by CoffeeScript 1.3.1
var BaseQuery, EventEmitter, assert, fluidize, toRelation,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; };

EventEmitter = require('events').EventEmitter;

toRelation = require('../nodes').toRelation;

assert = require('assert');

fluidize = require('../fluid');

module.exports = BaseQuery = (function(_super) {

  __extends(BaseQuery, _super);

  /*
    The base class for all queries. While this class itself is not part of
    gesundheits public API, the methods defined on it are.
  */


  BaseQuery.name = 'BaseQuery';

  BaseQuery.rootNode = null;

  function BaseQuery(engine, opts) {
    var table;
    if (opts == null) {
      opts = {};
    }
    /*
        :param engine: The engine the query will be bound to.
    
        :param opts.table:
          a ``String``, ``Relation``, ``Alias``, or object literal with a single
          key and value which will be interpreted as an alias name and table,
          respectively. This is normally given to as the first parameter to the
          query creation functions in :mod:`queries/index`
    */

    this.bind(engine);
    if (table = opts.table) {
      table = toRelation(table);
    }
    this.q = new this.constructor.rootNode(table);
  }

  BaseQuery.prototype.copy = function() {
    /* Instantiate a new query with a deep copy of this ones AST
    */

    var c;
    c = new this.constructor(this.engine);
    c.q = this.q.copy();
    c.bind(this.engine);
    return c;
  };

  BaseQuery.prototype.visit = function(fn) {
    /*
        Call the given function in the context of this query. This is useful with
        query factory functions where you can use it as a sort-of-DSL::
    
            SELECT('people', ['name'], function(q) {
              // this === q
              this.join('addresses', {
                on: {person_id: q.project('people', 'id')},
                fields: ['city', 'region']
              })
            })
    */
    return fn.call(this);
  };

  BaseQuery.prototype.bind = function(engine) {
    /*
        Bind this query object to a new engine.
        If no argument is given the query will be bound to the default engine.
    */

    var oldEngine, _ref, _ref1;
    oldEngine = this.engine;
    this.engine = engine || require('../').defaultEngine;
    if (this.engine !== oldEngine) {
      if (oldEngine != null ? oldEngine.unextend : void 0) {
        oldEngine.unextend(this);
      }
      if (this.engine.extend) {
        this.engine.extend(this);
      }
    }
    assert((_ref = this.engine) != null ? _ref.connect : void 0, "Engine has no connect method: " + this.engine);
    return assert((_ref1 = this.engine) != null ? _ref1.render : void 0, "Engine has no render method: " + this.engine);
  };

  BaseQuery.prototype.render = function() {
    /*
        Render the query to a SQL string.
    */
    return this.engine.render(this.q);
  };

  BaseQuery.prototype.compile = function() {
    /*
        Compile this query object, returning a SQL string and parameter array.
    */
    return [this.render(), this.q.params()];
  };

  BaseQuery.prototype.execute = function(cb) {
    /*
        Execute the query and buffer all results.
    
        :param cb: An (optional) node-style callback that will be called with any
          errors and/or the query results. If no callback is given, a new EventEmitter
          will be returned that emits either an 'error' or 'result' event.
    */

    var e,
      _this = this;
    if (!cb) {
      e = new EventEmitter;
      cb = function(err, res) {
        if (err) {
          return e.emit('error', err);
        } else {
          return e.emit('result', res);
        }
      };
    }
    this.engine.connect(function(err, conn) {
      if (err) {
        return cb(err);
      }
      return conn.query.apply(conn, _this.compile().concat([cb]));
    });
    return e;
  };

  BaseQuery.prototype.stream = function() {
    /*
        Execute the query, returning an EventEmitter that will stream the results.
    
        The exact events emitted depend on the underlying database engine. For
        example, MySQL query objects emit 'result' events, while Postgres query
        objects emit 'row' events.
    
        :param cb: A node-style callback that will be called with any errors and/or
          each row of the query results.
    */

    var e;
    e = new EventEmitter;
    this.engine.connect(function(err, conn) {
      var evt, q, _i, _len, _ref, _results;
      if (err) {
        return e.emit('error', err);
      }
      q = conn.query.apply(conn, this.compile());
      _ref = ['error', 'row', 'end', 'fields', 'result', 'close'];
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        evt = _ref[_i];
        _results.push(q.on(evt, this.emit.bind(e, evt)));
      }
      return _results;
    });
    return e;
  };

  BaseQuery.prototype.toString = function() {
    return this.render();
  };

  return BaseQuery;

})(EventEmitter);

fluidize(BaseQuery, 'bind', 'visit');
