// Generated by CoffeeScript 1.3.1
var Alias, And, JOIN_TYPES, Join, SUDQuery, Select, SelectQuery, fluidize, getAlias, sqlFunction, toRelation, _ref,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; },
  __slice = [].slice;

fluidize = require('../fluid');

SUDQuery = require('./sud');

_ref = require('../nodes'), Alias = _ref.Alias, getAlias = _ref.getAlias, Select = _ref.Select, And = _ref.And, Join = _ref.Join, toRelation = _ref.toRelation, sqlFunction = _ref.sqlFunction, JOIN_TYPES = _ref.JOIN_TYPES;

module.exports = SelectQuery = (function(_super) {

  __extends(SelectQuery, _super);

  /*
    Adds a number of SELECT-specific methods to :class:`queries/sud::SUDQuery`,
    such as `fields` and `groupBy`
  */


  SelectQuery.name = 'SelectQuery';

  function SelectQuery() {
    return SelectQuery.__super__.constructor.apply(this, arguments);
  }

  SelectQuery.rootNode = Select;

  SelectQuery.prototype.fields = function() {
    var alias, f, fields, project, rel, _i, _len, _results;
    fields = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /*
        Adds one or more fields to the query. If the second argument is an array, 
        the first argument is treated as a table (in the same way that :meth:`join` 
        understands tables) and the second argument as the list of fields to 
        select/update from that table. The table **must** already be joined for this
        to work.
    
        If the second argument is not an array, then each argument is treated as an 
        individual field to be projected from the currently focused table.
    */

    if (fields[1] && Array.isArray(fields[1])) {
      rel = this.q.relations.get(fields[0]);
      if (rel == null) {
        unknown('table', fields[0]);
      }
      fields = fields[1];
    } else {
      rel = this.q.relations.active;
    }
    if (fields.length === 0) {
      this.q.projections.prune(function(p) {
        return p.source === rel;
      });
      return;
    }
    project = function(f) {
      if (typeof f === 'object') {
        return f;
      } else {
        return rel.project(f);
      }
    };
    _results = [];
    for (_i = 0, _len = fields.length; _i < _len; _i++) {
      f = fields[_i];
      if (alias = getAlias(f)) {
        f = f[alias];
        _results.push(this.q.projections.addNode(new Alias(project(f), alias)));
      } else {
        _results.push(this.q.projections.addNode(project(f)));
      }
    }
    return _results;
  };

  SelectQuery.prototype.agg = function() {
    var alias, fields, fun, funcNode;
    fun = arguments[0], fields = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
    /*
        Adds one or more aggregated fields to the query
    
        :param fun: name of SQL aggregation function.
        :param fields: Fields to be projected from the current table and passed
          as arguments to ``fun``
    
        Example::
    
          select('t1').agg('count', 'id') # SELECT count(id) FROM t1
    */

    if (alias = getAlias(fun)) {
      fun = fun[alias];
    }
    funcNode = sqlFunction(fun, fields);
    if (alias) {
      return this.q.projections.addNode(new Alias(funcNode, alias));
    } else {
      return this.q.projections.addNode(funcNode);
    }
  };

  SelectQuery.prototype.distinct = function(bool) {
    /*
        Make this query DISTINCT on *all* fields.
    */
    return this.q.distinct.enable = bool;
  };

  SelectQuery.prototype.join = function(table, opts) {
    var joinClause, rel, type;
    if (opts == null) {
      opts = {};
    }
    /*
        Join another table to the query.
    
        :param table: A table name, or alias literal. An error will be thrown if
          the table/alias name is not unique. See :func:`nodes::toRelation` for
          more information on the many things ``table`` could be.
        :param opts.on:
          An object literal expressing join conditions. See
          :meth:`queries/select::SelectQuery::where` for more.
        :param opts.type: A join type constant (e.g. INNER, OUTER)
        :param opts.fields: A list of fields to be projected from the newly joined table.
    */

    rel = toRelation(table);
    if (this.q.relations.get(rel.ref(), false)) {
      throw new Error("Table/alias " + (rel.ref()) + " is not unique!");
    }
    type = opts.type || JOIN_TYPES.INNER;
    if (!(type instanceof JOIN_TYPES.INNER.constructor)) {
      throw new Error("Invalid join type " + type + ", try the constant types exported in the base module (e.g. INNER).");
    }
    joinClause = opts.on && new And(this.makeClauses(rel, opts.on));
    this.q.relations.addNode(new Join(type, rel, joinClause));
    this.q.relations.registerName(rel);
    this.q.relations["switch"](rel);
    if (opts.fields != null) {
      return this.fields.apply(this, opts.fields);
    }
  };

  SelectQuery.prototype.ensureJoin = function(table, opts) {
    var rel;
    if (opts == null) {
      opts = {};
    }
    /*
        The same as :meth:`join`, but will only join ``tbl`` if it is **not**
        joined already.
    */

    rel = toRelation(table);
    if (!this.q.relations.get(rel.ref(), false)) {
      return this.join(tbl, opts);
    }
  };

  SelectQuery.prototype.rel = function(alias) {
    /* A shorthand way to get a relation by (alias) name
    */
    return this.q.relations.get(alias);
  };

  SelectQuery.prototype.project = function(alias, field) {
    /*
        Return a :class:`nodes::Projection` node representing ``<alias>.<field>``.
    
        This node has a number methods from :class:`nodes::ComparableMixin` that can
        create new comparison nodes usable in join conditions and where clauses::
    
          # Find developers over the age of 45
          s = select('people', ['name'])
          dep_id = s.project('people', 'department_id')
          s.join('departments', on: {id: dep_id})
          s.where(s.project('departments', 'name').eq('development'))
          s.where(s.project('people', 'age').gte(45))
    */
    return this.q.relations.get(alias, true).project(field);
  };

  SelectQuery.prototype.focus = function(alias) {
    /*
        Make a different table "focused", this will use that table as the default
        for the ``fields``, ``order`` and ``where`` methods.
    
        :param alias: The table/alias name to focus. If the table or alias is not
          already part of the query an error will be thrown.
    */
    return this.q.relations["switch"](alias);
  };

  SelectQuery.prototype.groupBy = function() {
    var field, fields, rel, _i, _len, _results;
    fields = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /* Add a GROUP BY to the query.
    */

    rel = this.q.relations.active;
    _results = [];
    for (_i = 0, _len = fields.length; _i < _len; _i++) {
      field = fields[_i];
      if (field.constructor === String) {
        _results.push(this.q.groupBy.addNode(rel.project(field)));
      } else {
        _results.push(this.q.groupBy.addNode(field));
      }
    }
    return _results;
  };

  return SelectQuery;

})(SUDQuery);

fluidize(SelectQuery, 'distinct', 'fields', 'agg', 'join', 'ensureJoin', 'focus', 'groupBy');

SelectQuery.prototype.field = SelectQuery.prototype.fields;

SelectQuery.prototype.p = SelectQuery.prototype.project;
