// Generated by CoffeeScript 1.3.1
var BaseQuery, CONST_NODES, Or, OrderBy, SUDQuery, fluidize, nodes, unknown,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; },
  __slice = [].slice;

BaseQuery = require('./base');

nodes = require('../nodes');

fluidize = require('../fluid');

Or = nodes.Or, OrderBy = nodes.OrderBy, CONST_NODES = nodes.CONST_NODES;

module.exports = SUDQuery = (function(_super) {

  __extends(SUDQuery, _super);

  SUDQuery.name = 'SUDQuery';

  function SUDQuery() {
    return SUDQuery.__super__.constructor.apply(this, arguments);
  }

  /*
    SUDQuery is the base class for SELECT, UPDATE, and DELETE queries. It adds
    logic to :class:`queries/base::BaseQuery` for dealing with WHERE clauses and
    ordering.
  */


  SUDQuery.prototype.where = function(alias, predicate) {
    /*
        Add a WHERE clause to the query. Can optionally take a table/alias name as the
        first parameter, otherwise the clause is added using the last table added to
        the query.
    
        The where clause itself can be a comparison node, such as those produced by
        the :class:`nodes::ComparableMixin` methods::
    
          q.where(q.project('table','field1').eq(42))
          q.where(q.project('table','field2').gt(42))
    
        ... Or an object literal where each key is a field name (or field name
        alias) and each value is a constraint::
    
          q.where('table', {field1: 42, field2: {gt: 42}})
    
        Constraints values can also be other projected fields::
    
          p = q.project.bind(q, 'table')
          q.where('table', p('field1').gt(p('field2')))
    */

    var node, rel, _i, _len, _ref, _results;
    if (predicate != null) {
      rel = this.q.relations.get(alias);
      if (rel == null) {
        unknown('table', alias);
      }
    } else {
      predicate = alias;
      rel = this.defaultRel();
    }
    if (predicate.constructor !== Object) {
      return this.q.where.addNode(predicate);
    }
    _ref = this.makeClauses(rel, predicate);
    _results = [];
    for (_i = 0, _len = _ref.length; _i < _len; _i++) {
      node = _ref[_i];
      _results.push(this.q.where.addNode(node));
    }
    return _results;
  };

  SUDQuery.prototype.or = function() {
    var arg, args, clauses, rel, _i, _len;
    args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /*
        Add one or more WHERE clauses, all joined by the OR operator.
    */

    rel = this.defaultRel();
    clauses = [];
    for (_i = 0, _len = args.length; _i < _len; _i++) {
      arg = args[_i];
      clauses.push.apply(clauses, this.makeClauses(rel, arg));
    }
    return this.q.where.addNode(new Or(clauses));
  };

  SUDQuery.prototype.makeClauses = function(rel, constraint) {
    var clauses, field, op, predicate, val;
    clauses = [];
    for (field in constraint) {
      predicate = constraint[field];
      if (predicate === null) {
        clauses.push(rel.project(field).compare('IS', CONST_NODES.NULL));
      } else if (predicate.constructor === Object) {
        for (op in predicate) {
          val = predicate[op];
          clauses.push(rel.project(field).compare(op, val));
        }
      } else {
        clauses.push(rel.project(field).eq(predicate));
      }
    }
    return clauses;
  };

  SUDQuery.prototype.order = function() {
    var args, dir, direction, field, name, orderBy, orderings, rel, _i, _j, _len, _len1, _ref, _results;
    args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /*
        Add an ORDER BY to the query. Currently this *always* uses the "active"
        table of the query. (See :meth:`queries/select::SelectQuery.from`)
    
        Each ordering can either be a string, in which case it must be a valid-ish
        SQL snippet like 'some_field DESC', (the field name and direction will still
        be normalized) or an object, in which case each key will be treated as a
        field and each value as a direction.
    */

    rel = this.defaultRel();
    orderings = [];
    for (_i = 0, _len = args.length; _i < _len; _i++) {
      orderBy = args[_i];
      switch (orderBy.constructor) {
        case String:
          orderings.push(orderBy.split(' '));
          break;
        case OrderBy:
          this.q.orderBy.addNode(orderBy);
          break;
        case Object:
          for (name in orderBy) {
            dir = orderBy[name];
            orderings.push([name, dir]);
          }
          break;
        default:
          throw new Error("Can't turn " + orderBy + " into an OrderBy object");
      }
    }
    _results = [];
    for (_j = 0, _len1 = orderings.length; _j < _len1; _j++) {
      _ref = orderings[_j], field = _ref[0], direction = _ref[1];
      direction = (function() {
        switch ((direction || '').toLowerCase()) {
          case 'asc':
          case 'ascending':
            return 'ASC';
          case 'desc':
          case 'descending':
            return 'DESC';
          case '':
            return '';
          default:
            throw new Error("Unsupported ordering direction " + direction);
        }
      })();
      _results.push(this.q.orderBy.addNode(new OrderBy(rel.project(field), direction)));
    }
    return _results;
  };

  SUDQuery.prototype.limit = function(l) {
    /* Set the LIMIT on this query
    */
    return this.q.limit.value = l;
  };

  SUDQuery.prototype.offset = function(l) {
    /* Set the OFFSET of this query
    */
    return this.q.offset.value = l;
  };

  SUDQuery.prototype.defaultRel = function() {
    return this.q.relations.active;
  };

  return SUDQuery;

})(BaseQuery);

fluidize(SUDQuery, 'where', 'or', 'limit', 'offset', 'order');

unknown = function(type, val) {
  throw new Error("Unknown " + type + ": " + val);
};
